%{

/*
   pikts.l -- PIKT filter utility for stripping away comments

   $Header: /var/cvs/PIKT/pikt/src/piktc/pikts.l,v 1.23 2007/01/02 09:09:06 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   Usage:  pikts [-bGhlTVz]

           -b              strip away block comments only
           -G              run in debug mode
           -h              show program help
           -l              strip away line comments only
           -T              run in test mode
           -V              show version info
           -z              run in 'make check' mode

           note:  the default is to strip away both block and line comments

 */

#include "../lib/common.h"
#include "piktc.h"

const char pikts_version[] = "PIKT pikts version " VERSION;

int isdebug = FALSE;
int ishelp = FALSE;
int isshow_version = FALSE;
int istest = FALSE;
int isblockstrip = FALSE;
int islinestrip = FALSE;
int ismakecheck = FALSE;

int exstat = 0;

int bailout = FALSE;	/* signifies command-line error */

int piktslex(void);
void pikts_usage(void);
int isopt(char **, char **, int);

%}

W	[ \t]

%x CMNT0 CMNT1

%option noyywrap

%%

<INITIAL>("\\/*"|"/\\*"|"\\/\\*") { printf("/*"); }
<INITIAL>("\\*/"|"*\\/"|"\\*\\/") { printf("*/"); }
<INITIAL>("\\//"|"/\\/"|"\\/\\/") { printf("//"); }

<CMNT0,CMNT1>("\\/*"|"/\\*"|"\\/\\*") { ; }
<CMNT0,CMNT1>("\\*/"|"*\\/"|"\\*\\/") { ; }
<CMNT0,CMNT1>("\\//"|"/\\/"|"\\/\\/") { ; }

"/*"				{ if (isblockstrip)
					BEGIN(CMNT0);
				  else
					ECHO;
				}
{W}*"//"			{ if (islinestrip)
					BEGIN(CMNT1);
				  else
					ECHO;
				}
\n				{ ECHO; }
.				{ ECHO; }

<CMNT0>"*/"			{ BEGIN(INITIAL); }
<CMNT0>\n			{ ECHO; }
<CMNT0>.			{ ; }

<CMNT1>.+			{ ; }
<CMNT1>\n			{ BEGIN(INITIAL);
				  ECHO;
				}

%%

int
main (int argc, char *argv[])
{
	istest = isopt(argv+1, argv+argc-1, 'T');
	isdebug = isopt(argv+1, argv+argc-1, 'G');
	if((isshow_version = isopt(argv+1, argv+argc-1, 'V'))) {
		show_version(pikts_version);
		exstat = maxval(exstat, 0);
		bailout = TRUE;
	}
	if((ishelp = isopt(argv+1, argv+argc-1, 'h'))) {
		pikts_usage();
		exstat = maxval(exstat, 0);
		bailout = TRUE;
	}
	ismakecheck = isopt(argv+1, argv+argc-1, 'z');
	if (bailout)
		exit(exstat);
	isblockstrip = isopt(argv+1, argv+argc-1, 'b');
	islinestrip = isopt(argv+1, argv+argc-1, 'l');
	if (! isblockstrip && ! islinestrip) {	/* with no command-line args,
						   default to strip away all
						   comments
						 */
		isblockstrip = islinestrip = TRUE;
	}
	piktslex();
	exit(exstat);
}

/* output program usage to screen
 */
void
pikts_usage (void)
{
	printf("\nUsage:  pikts [-bGhlTV]\n\n");
	printf("\t-b\t\t\tstrip away block comments only\n");
	printf("\t-G\t\t\trun in debug mode\n");
	printf("\t-h\t\t\tshow program help\n");
	printf("\t-l\t\t\tstrip away line comments only\n");
	printf("\t-T\t\t\trun in test mode\n");
	printf("\t-V\t\t\tshow version info\n");
	printf("\n\tnote:  the default is to strip away both block and line comments\n\n");
	return;
}

/* look for a command-line (minus) switch, returning TRUE if found, else
   FALSE; this is for simpler command-line parsing; for more involved parsing,
   we use cmd.[lyc] and cmdparse()
 */
int
isopt (char **avcurr, char **avlast, int opt)
{
	while (avcurr <= avlast) {
		if (**avcurr == '-') {
			if (strchr(*avcurr, opt))
				return TRUE;
		}
		avcurr++;
	}
	return FALSE;
}
