%{

/*
   alt.y -- PIKT alarm script parser

   $Header: /var/cvs/PIKT/pikt/src/piktd/alt.y,v 1.48 2007/01/02 09:10:53 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   this is the sausage factory; not for the squeamish; enter at your own risk!

   the ifblk & elifs sections could probably be consolidated/abbreviated

 */

#include "../lib/common.h"
#include "piktd.h"

char _altparse_[] = "altparse";

#define DUMP(token)	if (isdebug) logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, _altparse_, FALSE, "@%s@", token)

int i, k;

int valtype = 0;
int dattype = NIL;

int doregvar = FALSE;

int vdt[MAXVAP];	/* var dat type */

int v;

int dim[MAXARY+1];	/* for nested arrays (e.g., $foo[#bar[1,2],3]), the
			   number of dimensions for the current array
			 */
int arynum;		/* the current array number */

char *nam;

stm *sp;

int ifnum, whilenum, repeatnum, fornum, switchnum;
int loopnum;

stm *loopsps[MAXLOOPS+3];	/* +3, because MAXLOOPS #defined as
				   3*MAXIFS in common.h
				 */

int isif[MAXIFS+1];
stm *ifsps[MAXIFS+1], *elifsps[MAXIFS+1], *elsesps[MAXIFS+1];
stm *elifsasps[MAXIFS+1], *elifszsps[MAXIFS+1];
stm *pelifsps[MAXIFS+1];
stm *stmszsps[MAXIFS+1], *fstmszsps[MAXIFS+1];

stm *whilesps[MAXWHILES+1], *endwhilesps[MAXWHILES+1];

stm *repeatsps[MAXREPEATS+1], *untilsps[MAXREPEATS+1];

stm *forsps[MAXFORS+1], *endforsps[MAXFORS+1];
stm *continuesps[MAXFORS+1];

stm *switchsps[MAXSWITCHS+1], *casesps[MAXSWITCHS+1], *defaultsps[MAXSWITCHS+1];
stm *lcasesps[MAXSWITCHS+1], *caseblkzsps[MAXSWITCHS+1];

stm *stmzsp;
stm *ifblkzsp, *whileblkzsp, *repeatblkzsp, *forblkzsp, *switchblkzsp;

stm *breaksp, *continuesp;

stm *dosp, *againsp, *leavesp, *redosp, *nextsp, *skipsp, *lastsp, *pausesp, *quitsp, *diesp, *outputsp, *setsp, *unsetsp, *execsp;

#define CHKARYNUM	if (arynum >= MAXARY) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_ARYNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXARY (%d) exceeded", ALMNAM, MAXARY); \
				STATUS = ERROR; \
			} \
			else if (arynum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_ARYNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, arynum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define CHKIFNUM	if (ifnum > MAXIFS) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_IFNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXIFS (%d) exceeded", ALMNAM, MAXIFS); \
				STATUS = ERROR; \
			} \
			else if (ifnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_IFNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, ifnum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define CHKWHILENUM	if (whilenum > MAXWHILES) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_WHILENUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXWHILES (%d) exceeded", ALMNAM, MAXWHILES); \
				STATUS = ERROR; \
			} \
			else if (whilenum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_WHILENUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, whilenum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define CHKREPEATNUM	if (repeatnum > MAXREPEATS) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_REPEATNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXREPEATS (%d) exceeded", ALMNAM, MAXREPEATS); \
				STATUS = ERROR; \
			} \
			else if (repeatnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_REPEATNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, repeatnum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define CHKFORNUM	if (fornum > MAXFORS) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_FORNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXFORS (%d) exceeded", ALMNAM, MAXFORS); \
				STATUS = ERROR; \
			} \
			else if (fornum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_FORNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, fornum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define CHKSWITCHNUM	if (switchnum > MAXSWITCHS) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_SWITCHNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXSWITCHS (%d) exceeded", ALMNAM, MAXSWITCHS); \
				STATUS = ERROR; \
			} \
			else if (switchnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_SWITCHNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, switchnum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define CHKLOOPNUM	if (loopnum > MAXLOOPS) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_LOOPNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXLOOPS (%d) exceeded", ALMNAM, MAXLOOPS); \
				STATUS = ERROR; \
			} \
			else if (loopnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_LOOPNUM_LESS_THAN_ZERO, __FILE__, __LINE__, _altparse_, FALSE, "in %s, loopnum < 0", ALMNAM); \
				STATUS = ERROR; \
			}

#define ENTERIF		ifnum++;  \
			if (ifnum > MAXIFS) { \
				logreturn(PIKTLOG, LOG_ERR, ERRID_IFNUM_MAX_EXCEEDED, __FILE__, __LINE__, _altparse_, FALSE, "in %s, MAXIFS (%d) exceeded", ALMNAM, MAXIFS); \
				STATUS = ERROR; \
			}

#define EXITIF		isif[ifnum] = 0; \
			ifblkzsp = ifsps[ifnum]->aux; \
			ifsps[ifnum] = elifsps[ifnum] = elsesps[ifnum] = elifsasps[ifnum] = elifszsps[ifnum] = pelifsps[ifnum] = stmszsps[ifnum] = fstmszsps[ifnum] = (stm *) NULL; \
			ifnum--

#define ENTERWHILE	whilenum++; \
			loopnum++; \
			CHKWHILENUM;

#define EXITWHILE	whileblkzsp = whilesps[whilenum]->aux; \
			whilesps[whilenum] = endwhilesps[whilenum] = (stm *) NULL; \
			loopsps[loopnum] = (stm *) NULL; \
			whilenum--; \
			loopnum--

#define ENTERREPEAT	repeatnum++; \
			loopnum++; \
			CHKREPEATNUM;

#define EXITREPEAT	repeatblkzsp = repeatsps[repeatnum]->aux; \
			repeatsps[repeatnum] = untilsps[repeatnum] = (stm *) NULL; \
			loopsps[loopnum] = (stm *) NULL; \
			repeatnum--; \
			loopnum--


#define ENTERFOR	fornum++; \
			loopnum++; \
			CHKFORNUM;

#define EXITFOR		forblkzsp = forsps[fornum]->aux; \
			forsps[fornum] = endforsps[fornum] = continuesps[fornum] = (stm *) NULL; \
			loopsps[loopnum] = (stm *) NULL; \
			fornum--; \
			loopnum--

#define ENTERSWITCH	switchnum++; \
			CHKSWITCHNUM;

#define EXITSWITCH	switchblkzsp = switchsps[switchnum]->aux; \
			switchsps[switchnum] = casesps[switchnum] = defaultsps[switchnum] = lcasesps[switchnum] = caseblkzsps[switchnum] = (stm *) NULL; \
			switchnum--

/* set the queued cases tnext to endswitch */
#define LINK_QUEUED_CASES_TO_ENDSW \
			sp = switchsps[switchnum]->aux->aux; \
			while (sp != casesps[switchnum]) { \
				sp->tnext = switchsps[switchnum]->aux; \
				sp = sp->next; \
			} \
			sp->tnext = switchsps[switchnum]->aux

/* set the queued cases tnext to the first case statement */
#define LINK_QUEUED_CASES_TO_CURR_CASE \
			sp = switchsps[switchnum]->aux->aux; \
			while (sp != casesps[switchnum]) { \
				sp->tnext = casesps[switchnum]->tnext; \
				sp = sp->next; \
			}

/* set the queued cases tnext to the first default statement */
#define LINK_QUEUED_CASES_TO_CURR_DEFAULT \
			sp = switchsps[switchnum]->aux->aux; \
			while (sp != casesps[switchnum]) { \
				sp->tnext = defaultsps[switchnum]->tnext; \
				sp = sp->next; \
			} \
			sp->tnext = defaultsps[switchnum]->tnext

/* set the last-statement tnext in the previous caseblk to the first
   statement in the current caseblk
 */
#define LINK_PREV_CASE_TO_CURR_CASE \
			if (caseblkzsps[switchnum]) \
				caseblkzsps[switchnum]->tnext = lcasesps[switchnum]->tnext

/* set the last-statement tnext in the previous caseblk to the first
   statement in the current defaultblk
 */
#define LINK_PREV_CASE_TO_CURR_DEFAULT \
			if (caseblkzsps[switchnum]) \
				caseblkzsps[switchnum]->tnext = defaultsps[switchnum]->tnext

%}

%union {
	int ival;
	char *sval;
}

%token <sval> _STANZA
%token _STATUS _ACTIVE _INACTIVE _SUSPENDED _TESTING _DEBUGGING
%token _LEVEL _EMERG _ALERT _CRIT _ERR _WARNING _NOTICE _INFO _DEBUG
%token _TASK
%token _INPUT _PROCESS _REGFILE _LOGFILE
%token _FILTER
%token _SEPARATORS _DAT _KEYS
%token _INIT
%token _BEGIN _END
%token _RULE
%token _DO
%token _IF _ELIF _ELSE _ENDIF
%token _WHILE _ENDWHILE _REPEAT _UNTIL _FOR _FOREACH _ENDFOR
%token _BREAK _CONTINUE
%token _SWITCH _CASE _DEFAULT _BREAKSWITCH _ENDSWITCH
%token _SWITCHLP _FORLP
%token _AGAIN _LEAVE _REDO _NEXT _SKIP _LAST _PAUSE _QUIT _DIE
%token _OUTPUT _MAIL _PRINT _PIKTLOG _SYSLOG
%token _SET _UNSET _EXEC _WAIT
%token <sval> _STR _STRVAR
%token <sval> _POSITION _RANGE
%token <sval> _VARNUM _VARSTR _VARHSTSTR _VARHSTNUM _VARPRCSTR _VARPRCNUM
%token <sval> _FLHNDL
%token <sval> _VALNUM _VALSTR
%token <sval> _NOP _SOP _FOP _NREL _SREL
%token <sval> _NUM _INT
%token <sval> _FNNUM _FNSTR
%token _NOT _AND _OR

%type <sval> position varnum varstr varnow
%type <ival> exp ndx
%type <ival> ndxnums

%left _OR
%left _AND
%nonassoc _NREL _SREL
%left '+' '-' _NOP _SOP
%nonassoc _FOP
%nonassoc _NOT '~' _UNOP

%expect 2

%%

start:		alarmdefs
		| scriptdef
		;

alarmdefs:	alarmdef
		| alarmdefs alarmdef
		;

alarmdef:	stanza scriptdef
		;

scriptdef:	begin
		| end
		| begin end
		| init begin
		| init end
		| init begin end
		| init dataspec begin
		| init dataspec end
		| init dataspec begin end
		| init dataspec rules
		| init dataspec begin rules
		| init dataspec rules end
		| init dataspec begin rules end
		| statements
		;

stanza:		_STANZA {
			if (scanpass == 1) {
				regalm($1);
			} else if (scanpass == 2) {
				DUMP($1);
				if (! alarmcurr)
					alarmcurr = alarms;
				else
					alarmcurr = alarmcurr->next;
				free($1);
				INPUTTYP = NOINPUT;	/* initialization;
							   possibly override
							   later in dataspec
							 */
			}
		}
		;

init:		_INIT {
			if (scanpass == 2) {
				if (! interactive)
					logexit(PIKTLOG, LOG_ERR, ERRID_STATUS_LEVEL_TASK_MISSING, __FILE__, __LINE__, _altparse_, FALSE, "in %s, missing status, level & task", ALMNAM);
			}
		}
		| _INIT status level task
		;

status:		_STATUS _ACTIVE { if (scanpass == 2) STATUS = ACTIVE; }
		| _STATUS _INACTIVE { if (scanpass == 2) STATUS = INACTIVE; }
		| _STATUS _SUSPENDED { if (scanpass == 2) STATUS = SUSPENDED; }
		| _STATUS _TESTING { if (scanpass == 2) STATUS = TESTING; }
		| _STATUS _DEBUGGING { if (scanpass == 2) STATUS = DEBUGGING; }
		;
		
level:		_LEVEL _EMERG { if (scanpass == 2) LEVEL = LOG_EMERG; }
		| _LEVEL _ALERT { if (scanpass == 2) LEVEL = LOG_ALERT; }
		| _LEVEL _CRIT { if (scanpass == 2) LEVEL = LOG_CRIT; }
		| _LEVEL _ERR { if (scanpass == 2) LEVEL = LOG_ERR; }
		| _LEVEL _WARNING { if (scanpass == 2) LEVEL = LOG_WARNING; }
		| _LEVEL _NOTICE { if (scanpass == 2) LEVEL = LOG_NOTICE; }
		| _LEVEL _INFO { if (scanpass == 2) LEVEL = LOG_INFO; }
		| _LEVEL _DEBUG { if (scanpass == 2) LEVEL = LOG_DEBUG; }
		;

task:		_TASK _STR {
			if (scanpass == 2) {
				dequote($2);
				DUMP($2);
				FREE(TASK);
				TASK = $2;
			} else
				free($2);
		}
		;

dataspec:	input filter separators inputdata keys
		;

input:		_INPUT _STR {
			if (scanpass == 2) {
				DUMP($2);
				INPUTTYP = PROCINPUT;
				INPUTSRC = (char *) mymalloc((size_t)(strlen($2)+8) * sizeof(char), _altparse_);
				sprintf(INPUTSRC, "echo \'%s\'", $2);
				free($2);
			} else
				free($2);
		}
		| _INPUT _PROCESS _STR {
			if (scanpass == 2) {
				DUMP($3);
				INPUTTYP = PROCINPUT;
				INPUTSRC = $3;
			} else
				free($3);
		}
		| _INPUT _REGFILE _STR {
			if (scanpass == 2) {
				DUMP($3);
				INPUTTYP = FILEINPUT;
				INPUTSRC = $3;
			} else
				free($3);
		}
		| _INPUT _REGFILE _STR _STR {
			if (scanpass == 2) {
				DUMP($3);
				INPUTTYP = FILEINPUT;
				INPUTSRC = $3;
				INPUTTMPFIL = (char *) mymalloc((size_t)(strlen($3)+9) * sizeof(char), _altparse_);
				sprintf(INPUTTMPFIL, "%s.pikttmp", $3);
				INPUTBAKFIL = $4;
			} else {
				/* INPUTBAKFIL needs to be set also in
				   scanpass 1 in case we call varndx() on
				   outlin, outline, or outputline; regalm()
				   is called earlier in scanpass 1, so it's
				   safe to assign to INPUTBAKFIL here
				 */
				INPUTBAKFIL = $4;
				free($3);
				free($4);
			}
		}
		| _INPUT _LOGFILE _STR {
			if (scanpass == 2) {
				DUMP($3);
				INPUTTYP = LOGINPUT;
				INPUTSRC = $3;
				INPUTLOGNAME = $3;
				dohistory = TRUE;	/* so, process history file */
				HSTTYP |= LOGINPUT;
			} else
				free($3);
		}
		;

filter:		/* empty */ { ; }
		| _FILTER _STR {
			if (scanpass == 2) {
				DUMP($2);
				FILTER = $2;
			} else
				free($2);
		}
		;

separators:	/* empty */ {
			if (scanpass == 2) {
				/* space/tab */
				SEPARATORS = mystrdup(" \t", _altparse_);
			}
		}
		| _SEPARATORS _STR {
			if (scanpass == 2) {
				DUMP($2);
				SEPARATORS = $2;
			} else
				free($2);
		}
		;

inputdata:	/* empty */
		| data
		;

data:		datum
		| data datum
		;

datum:		_DAT varnow '[' _INT ']' position {
			valtype = FIXEDVAL;
			if (scanpass == 2) {
				DUMP($2);
				DUMP("[");
				DUMP($4);
				DUMP("]");
				DUMP($6);
			} else {
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				errno = 0;
				arndx[1] = atol($4);
				if (errno || (arndx[1] > MAXARNDX)) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_ARRAY_INDEX_INVALID, __FILE__, __LINE__, _altparse_, FALSE, "in %s, bad or illegal array index: [%s]", ALMNAM, $4);
					STATUS = ERROR;
					for (k=0; k<=MAXARYDIM; k++)
						arndx[k] = NILARNDX;
				}
				regvar($2+1, valtype, dattype, NUMBER, arndx[1] > 0 ? 1 : 0, arndx, (char *)NULL, $6);
			}
			free($2);
			free($4);
			free($6);
		}
		| _DAT varnow '[' _STR ']' position {
			valtype = FIXEDVAL;
			if (scanpass == 2) {
				DUMP($2);
				DUMP("[");
				DUMP($4);
				DUMP("]");
				DUMP($6);
			} else {
				/* pass arndxnil instead
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = NILARNDX;
				 */
				regvar($2+1, valtype, dattype, STRING, 1, arndxnil, mystrdup($4,_altparse_), $6);
			}
			free($2);
			free($4);
			free($6);
		}
		| _DAT _STR {
			if (scanpass == 2) {
				POSTYP = REGEXPR;
				REGEXP = mystrdup(stripesc($2), _altparse_);
			}
			free($2);
		}
		;

position:	_POSITION {
			/* DUMP($1); dumped in datum above */
			$$ = $1;
		}
		| _RANGE {
			DUMP($1);
			$$ = $1;
		}
		;

keys:		/* empty */ { ;	}
		| _KEYS keyset
		;

keyset:		varnowset
		| keyset varnowset
		;

varnowset:	varnow '[' _INT ']' {
			if (scanpass == 2) {
				DUMP($1);
				DUMP("[");
				DUMP($3);
				DUMP("]");
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				arndx[1] = atol($3);
				regkey($1, FIXEDVAL|FREEVAL, arndx, (char *)NULL);
			}
			free($1);
			free($3);
		}
		| varnow '[' _STR ']' {
			if (scanpass == 2) {
				DUMP($1);
				DUMP("[");
				DUMP($3);
				DUMP("]");
				/* pass arndxnil instead
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = NILARNDX;
				 */
				regkey($1, FIXEDVAL|FREEVAL, arndxnil, $3);
			}
			free($1);
			free($3);
		}
		;

varnow:		_VARNUM { dattype = NUMBER; $$ = $1; }
		| _VARSTR { dattype = STRING; $$ = $1; }
		;

begin:		_BEGIN {
			valtype = BEGINVAL;
			if (scanpass == 2) {
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
				regrul(BEG);
			}
		}
		statements {
			#if DEEPDEBUG
			if (scanpass == 2) {
				if (isdebug)
					dumpsps(alarmcurr->begin->statements, "leaving _BEGIN");
			}
			#endif
			;
		}
		;

end:		_END {
			valtype = ENDVAL;
			if (scanpass == 2) {
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
				regrul(END);
			}
		}
		statements {
			#if DEEPDEBUG
			if (scanpass == 2) {
				if (isdebug)
					dumpsps(alarmcurr->end->statements, "leaving _END");
			}
			#endif
			;
		}
		;

rules:		rule
		| rules rule
		;

rule:		_RULE {
			valtype = FREEVAL;
			if (scanpass == 2) {
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
				regrul(RUL);
			}
		}
		statements {
			#if DEEPDEBUG
			if (scanpass == 2) {
				if (isdebug)
					dumpsps(alarmcurr->rules->statements, "leaving _RULE");
			}
			#endif
			;
		}
		;

statements:	statement {
			/* valtype is still 0 (its initialization value)
			   if we haven't run through begin, end, or rule--
			   i.e., this is a script consisting only of
			   statements
			 */
			if (! valtype) {
				regrul(END);
				valtype = ENDVAL;
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
			}
			if (scanpass == 2) {
				if (ifnum >= 0) {
					stmszsps[ifnum] = stmzsp;
					if (isif[ifnum])
						fstmszsps[ifnum] = stmzsp;
				}
			}
		}
		| statements statement {
			if (scanpass == 2) {
				if (ifnum >= 0) {
					stmszsps[ifnum] = stmzsp;
					if (isif[ifnum])
						fstmszsps[ifnum] = stmzsp;
				}
			}
		}
		;

statement:	do { if (scanpass == 2) stmzsp = dosp; }
		| ifblk { if (scanpass == 2) stmzsp = ifblkzsp; }
		| whileblk { if (scanpass == 2) stmzsp = whileblkzsp; }
		| repeatblk { if (scanpass == 2) stmzsp = repeatblkzsp; }
		| forblk { if (scanpass == 2) stmzsp = forblkzsp; }
		| switchblk { if (scanpass == 2) stmzsp = switchblkzsp; }
		| break { if (scanpass == 2) stmzsp = breaksp; }
		| continue { if (scanpass == 2) stmzsp = continuesp; }
		| again { if (scanpass == 2) stmzsp = againsp; }
		| leave { if (scanpass == 2) stmzsp = leavesp; }
		| redo { if (scanpass == 2) stmzsp = redosp; }
		| next { if (scanpass == 2) stmzsp = nextsp; }
		| skip { if (scanpass == 2) stmzsp = skipsp; }
		| last { if (scanpass == 2) stmzsp = lastsp; }
		| pause { if (scanpass == 2) stmzsp = pausesp; }
		| quit { if (scanpass == 2) stmzsp = quitsp; }
		| die { if (scanpass == 2) stmzsp = diesp; }
		| output { if (scanpass == 2) stmzsp = outputsp; }
		| set { if (scanpass == 2) stmzsp = setsp; }
		| unset { if (scanpass == 2) stmzsp = unsetsp; }
		| exec { if (scanpass == 2) stmzsp = execsp; }
		;

do:		_DO {
			if (scanpass == 2) {
				regstm(DO);
				dosp = stmcurr;
			}
		} exp
		;

ifblk:		if endif {
			if (scanpass == 2) {
				EXITIF;
			}
		}
		| if statements endif {
			if (scanpass == 2) {
				EXITIF;
			}
		}
		| if else endif {
			if (scanpass == 2) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements else endif {
			if (scanpass == 2) {
				ifsps[ifnum]->fnext = elsesps[ifnum];
				stmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if else statements endif {
			if (scanpass == 2) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements else statements endif {
			if (scanpass == 2) {
				ifsps[ifnum]->fnext = elsesps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if elifs endif {
			if (scanpass == 2) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->fnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if statements elifs endif {
			if (scanpass == 2) {
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->fnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if elifs else endif {
			if (scanpass == 2) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements elifs else endif {
			if (scanpass == 2) {
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if elifs else statements endif {
			if (scanpass == 2) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements elifs else statements endif {
			if (scanpass == 2) {
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		;

elifs:		elif {
			if (scanpass == 2) {
				pelifsps[ifnum] = elifsps[ifnum];
				elifsasps[ifnum] = elifsps[ifnum];
				elifszsps[ifnum] = elifsps[ifnum];
			}
		}
		| elif statements {
			if (scanpass == 2) {
				elifsps[ifnum]->fnext = ifsps[ifnum]->aux;
				pelifsps[ifnum] = elifsps[ifnum];
				elifsasps[ifnum] = elifsps[ifnum];
				elifszsps[ifnum] = stmszsps[ifnum];
			}
		}
		| elifs elif {
			if (scanpass == 2) {
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elifsps[ifnum];
				pelifsps[ifnum] = elifsps[ifnum];
				elifszsps[ifnum] = elifsps[ifnum];
			}
		}
		| elifs elif statements {
			if (scanpass == 2) {
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elifsps[ifnum];
				pelifsps[ifnum] = elifsps[ifnum];
				elifsps[ifnum]->fnext = ifsps[ifnum]->aux;
				elifszsps[ifnum] = stmszsps[ifnum];
			}
		}
		;

if:		_IF {
			if (scanpass == 2) {
				ENTERIF;
				regstm(IF);
				ifsps[ifnum] = stmcurr;
				isif[ifnum] = TRUE;
			}
		}
		conditions
		;

elif:		_ELIF {
			if (scanpass == 2) {
				regstm(ELIF);
				elifsps[ifnum] = stmcurr;
				isif[ifnum] = FALSE;
			}
		}
		conditions
		;

else:		_ELSE {
			if (scanpass == 2) {
				regstm(ELSE);
				elsesps[ifnum] = stmcurr;
				isif[ifnum] = FALSE;
			}
		}
		;

endif:		_ENDIF {
			if (scanpass == 2) {
				stmcurr->next = stmcurr->tnext = ifsps[ifnum]->aux;
				stmcurr = ifsps[ifnum]->aux;
				isif[ifnum] = FALSE;
			}
		}
		;

whileblk:	while endwhile {
			if (scanpass == 2) {
				endwhilesps[whilenum]->tnext = whilesps[whilenum];
				EXITWHILE;
			}
		}
		| while statements endwhile {
			if (scanpass == 2) {
				endwhilesps[whilenum]->tnext = whilesps[whilenum];
				EXITWHILE;
			}
		}
		;

while:		_WHILE {
			if (scanpass == 2) {
				ENTERWHILE;
				regstm(WHILE);
				whilesps[whilenum] = loopsps[loopnum] = stmcurr;
			}
		}
		conditions
		;

endwhile:	_ENDWHILE {
			if (scanpass == 2) {
				regstm(ENDWHILE);
				endwhilesps[whilenum] = stmcurr;
				stmcurr->next = whilesps[whilenum]->aux;
				stmcurr = whilesps[whilenum]->aux;
			}
		}
		;

repeatblk:	repeat until {
			if (scanpass == 2) {
				untilsps[repeatnum]->fnext = repeatsps[repeatnum];
				EXITREPEAT;
			}
		}
		| repeat statements until {
			if (scanpass == 2) {
				untilsps[repeatnum]->fnext = repeatsps[repeatnum];
				EXITREPEAT;
			}
		}
		;

repeat:		_REPEAT {
			if (scanpass == 2) {
				ENTERREPEAT;
				regstm(REPEAT);
				repeatsps[repeatnum] = loopsps[loopnum] = stmcurr;
			}
		}
		;

until:		_UNTIL {
			if (scanpass == 2) {
				regstm(UNTIL);
				untilsps[repeatnum] = stmcurr;
			}
		}
		conditions {
			if (scanpass == 2) {
				stmcurr->next = stmcurr->tnext = repeatsps[repeatnum]->aux;
				stmcurr = repeatsps[repeatnum]->aux;
			}
		}
		;

forblk:		for endfor {
			if (scanpass == 2) {
				endforsps[fornum]->tnext = forsps[fornum];
				EXITFOR;
			}
		}
		| for statements endfor {
			if (scanpass == 2) {
				endforsps[fornum]->tnext = forsps[fornum];
				EXITFOR;
			}
		}
		| foreach endfor {
			if (scanpass == 2) {
				endforsps[fornum]->tnext = forsps[fornum];
				EXITFOR;
			}
		}
		| foreach statements endfor {
			if (scanpass == 2) {
				endforsps[fornum]->tnext = forsps[fornum];
				EXITFOR;
			}
		}
		;

for:		_FOR forexps
		| _FORLP forexps ')'
		;

forexps:	_VARNUM '[' _INT ']' '=' {
			if (scanpass == 2) {
				ENTERFOR;
				DUMP($1);
				regstm(SET);
			}
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			errno = 0;
			arndx[1] = atol($3);
			if (errno || (arndx[1] > MAXARNDX)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_ARRAY_INDEX_INVALID, __FILE__, __LINE__, _altparse_, FALSE, "in %s, bad or illegal array index: [%s]", ALMNAM, $3);
				STATUS = ERROR;
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = NILARNDX;
			}
			regvar($1+1, valtype, NUMBER, NUMBER, arndx[1] > 0 ? 1 : 0, arndx, (char *)NULL, (char *)NULL);
			free($1);
			free($3);
		} expnum {
			if (scanpass == 2) {
				regstm(FOR);
				forsps[fornum] = loopsps[loopnum] = stmcurr;
			}
		} conditions
		_VARNUM '[' _INT ']' '=' {
			if (scanpass == 2) {
				stmcurr = forsps[fornum]->aux;
				DUMP($10);
			}
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			errno = 0;
			arndx[1] = atol($12);
			if (errno || (arndx[1] > MAXARNDX)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_ARRAY_INDEX_INVALID, __FILE__, __LINE__, _altparse_, FALSE, "in %s, bad or illegal array index: [%s]", ALMNAM, $12);
				STATUS = ERROR;
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = NILARNDX;
			}
			regvar($10+1, valtype, NUMBER, NUMBER, arndx[1] > 0 ? 1 : 0, arndx, (char *)NULL, (char *)NULL);
			free($10);
			free($12);
		} expnum {
			if (scanpass == 2) {
				stmcurr = forsps[fornum];
			}
		}
		;

foreach:	_FOREACH _FNNUM '(' varstr '[' _INT ']' ',' {
			nam = (char *) mymalloc((strlen($2+1)+3) * sizeof(char), _altparse_);
			sprintf(nam, " %s ", $2+1);
			if (! strstr(" split parse members keys skeys sortkeys rkeys reversekeys ", nam)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FOREACH_FUNCTION_INVALID, __FILE__, __LINE__, _altparse_, FALSE, "in %s, #%s() is not an approved foreach function", ALMNAM, $2);
				STATUS = ERROR;
			}
			free(nam);
			/* fake an index var, e.g., #i_f */
			nam = (char *) mymalloc((strlen($4+1)+4) * sizeof(char), _altparse_);
			sprintf(nam, "#i_%s", $4+1);
			if (scanpass == 2) {
				ENTERFOR;
				DUMP($2);
				DUMP($4);
				regstm(SET);
			}
			/* pass arndxnul instead
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			 */
			regvar(nam+1, valtype, NUMBER, NUMBER, 0, arndxnul, (char *)NULL, (char *)NULL);
			if (scanpass == 2) {
				regtok(mystrdup("1",_altparse_), FALSE);
				regstm(FOR);
				forsps[fornum] = loopsps[loopnum] = stmcurr;
			}
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			arndx[1] = (signed long) atol($6);
			regvar($4+1, valtype, STRING, NUMBER, 1, arndx, (char *)NULL, (char *)NULL);
			LISTINDEX(varndx($4+1)) = varndx(nam+1);
			if (scanpass == 2) {
				regtok(mystrdup(nam,_altparse_), FALSE);
				regtok(mystrdup("[",_altparse_), FALSE);
				regtok(mystrdup("0",_altparse_), FALSE);
				regtok(mystrdup("]",_altparse_), FALSE);
				regtok(mystrdup("<=",_altparse_), FALSE);
				regtok($2, TRUE);
				regtok(mystrdup("(", _altparse_), FALSE);
				regtok(mystrdup("\"",_altparse_), FALSE);
				regtok($4, FALSE);
				regtok(mystrdup("[",_altparse_), FALSE);
				regtok(mystrdup("0",_altparse_), FALSE);
				regtok(mystrdup("]",_altparse_), FALSE);
				regtok(mystrdup("\"",_altparse_), FALSE);
				regtok(mystrdup(",", _altparse_), FALSE);
			}
		}
		args ')' {
			if (scanpass == 2) {
				regtok(mystrdup(")", _altparse_), FALSE);
				stmcurr = forsps[fornum]->aux;
			}
			/* pass arndxnul instead
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = 0L;
			 */
			regvar(nam+1, valtype, NUMBER, NUMBER, 0, arndxnul, (char *)NULL, (char *)NULL);
			if (scanpass == 2) {
				regtok(mystrdup(nam,_altparse_), FALSE);
				regtok(mystrdup("[",_altparse_), FALSE);
				regtok(mystrdup("0",_altparse_), FALSE);
				regtok(mystrdup("]",_altparse_), FALSE);
				regtok(mystrdup("+",_altparse_), FALSE);
				regtok(mystrdup("1",_altparse_), FALSE);
				stmcurr = forsps[fornum];
			}
			free($6);
			free(nam);
		}
		;

endfor:		_ENDFOR {
			if (scanpass == 2) {
				regstm(SET);
				stmcurr->rhsexpr = forsps[fornum]->aux->rhsexpr;
				stmcurr->vrndx = forsps[fornum]->aux->vrndx;
				forsps[fornum]->aux->rhsexpr = (tok *) NULL;
				if (continuesps[fornum]) {
					continuesps[fornum]->fnext = stmcurr;
					continuesps[fornum] = (stm *) NULL;
				}
				regstm(ENDFOR);
				endforsps[fornum] = stmcurr;
				stmcurr->next = forsps[fornum]->aux;
				stmcurr = forsps[fornum]->aux;
			}
		}
		;

switchblk:	switch cases endswitch {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_ENDSW;
				EXITSWITCH;
			}
		}
		| switch caseblks cases endswitch {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_ENDSW;
				EXITSWITCH;
			}
		}
		| switch caseblks endswitch {
			if (scanpass == 2) {
				EXITSWITCH;
			}
		}
		| switch caseblks defaultblk endswitch {
			if (scanpass == 2) {
				/* set fnext for previous case to default */
				if (lcasesps[switchnum])
					lcasesps[switchnum]->fnext = defaultsps[switchnum];
				EXITSWITCH;
			}
		}
		| switch defaultblk endswitch {
			if (scanpass == 2) {
				EXITSWITCH;
			}
		}
		;

caseblks:	caseblk
		| caseblks caseblk
		;

caseblk:	cases statements {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_CURR_CASE;
				LINK_PREV_CASE_TO_CURR_CASE;
				/* note last statement in this caseblk */
				caseblkzsps[switchnum] = stmzsp;
			}
		}
		| cases breakswitch {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_ENDSW;
				caseblkzsps[switchnum] = (stm *) NULL;
			}
		}
		| cases statements breakswitch {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_CURR_CASE;
				LINK_PREV_CASE_TO_CURR_CASE;
				caseblkzsps[switchnum] = (stm *) NULL;
			}
		}
		;

defaultblk:	default
		| default statements {
			if (scanpass == 2) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
			}
		}
		| default breakswitch
		| default statements breakswitch {
			if (scanpass == 2) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
			}
		}
		| cases default {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_ENDSW;
			}
		}
		| cases default statements {
			if (scanpass == 2) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
				LINK_QUEUED_CASES_TO_CURR_DEFAULT;
			}
		}
		| cases default breakswitch {
			if (scanpass == 2) {
				LINK_QUEUED_CASES_TO_ENDSW;
			}
		}
		| cases default statements breakswitch {
			if (scanpass == 2) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
				LINK_QUEUED_CASES_TO_CURR_DEFAULT;
			}
		}
		;

cases:		case {
			if (scanpass == 2) {
				/* note first case in series of cases in endswitch aux */
				switchsps[switchnum]->aux->aux = casesps[switchnum];
				/* set fnext for previous case, if any, to this case */
				if (lcasesps[switchnum])
					lcasesps[switchnum]->fnext = casesps[switchnum];
				/* note latest case in series of cases */
				lcasesps[switchnum] = casesps[switchnum];
			}
		}
		| cases case {
			if (scanpass == 2) {
				/* note latest case in series of cases */
				lcasesps[switchnum] = casesps[switchnum];
			}
		}
		;

switch:		_SWITCH switchexp
		| _SWITCHLP switchexp ')'
		;

switchexp:	{
			if (scanpass == 2) {
				ENTERSWITCH;
				regstm(SWITCH);
				switchsps[switchnum] = stmcurr;
			}
		} exp {
			if (scanpass == 2) {
				regtok(mystrdup($2 == STRING ? "eq" : "==", _altparse_), FALSE);
			}
		}
		;	

case:		_CASE {
			if (scanpass == 2) {
				regstm(CASE);
				casesps[switchnum] = stmcurr;
				duptok(switchsps[switchnum]->rhsexpr, &stmcurr->rhsexpr);
				casesps[switchnum]->fnext = switchsps[switchnum]->aux;
			}
		} exp
		;

default:	_DEFAULT {
			if (scanpass == 2) {
				regstm(DEFAULT);
				defaultsps[switchnum] = stmcurr;
			}
		}
		;

breakswitch:	_BREAKSWITCH {
			if (scanpass == 2) {
				regstm(BREAKSWITCH);
				stmcurr->fnext = switchsps[switchnum]->aux;
			}
		}
		;

endswitch:	_ENDSWITCH {
			if (scanpass == 2) {
				stmcurr->next = stmcurr->tnext = switchsps[switchnum]->aux;
				stmcurr = switchsps[switchnum]->aux;
			}
		}
		;

break:		_BREAK {
			if (scanpass == 2) {
				if (loopnum == 0)
					logreturn(PIKTLOG, LOG_ERR, ERRID_BREAK_WITHOUT_LOOP, __FILE__, __LINE__, _altparse_, FALSE, "in %s, break with no corresponding loop", ALMNAM);
				else
					CHKLOOPNUM;
				regstm(BREAK);
				stmcurr->fnext = loopsps[loopnum]->aux;
				breaksp = stmcurr;
			}
		}
		;

continue:	_CONTINUE {
			if (scanpass == 2) {
				if (loopnum == 0)
					logreturn(PIKTLOG, LOG_ERR, ERRID_CONTINUE_WITHOUT_LOOP, __FILE__, __LINE__, _altparse_, FALSE, "in %s, continue with no corresponding loop", ALMNAM);
				else
					CHKLOOPNUM;
				regstm(CONTINUE);
				if (loopsps[loopnum]->type == FOR) {	/* a for loop */
					if (! continuesps[fornum])	/* first continue */
						continuesps[fornum] = stmcurr;
					else	/* subsequent continue, link to first continue */
						stmcurr->fnext = continuesps[fornum];
				} else {	/* not a for loop */
					stmcurr->fnext = loopsps[loopnum];
				}
				continuesp = stmcurr;
			}
		}
		;

again:		_AGAIN {
			if (scanpass == 2) {
				regstm(AGAIN);
				againsp = stmcurr;
			}
		}
		;

leave:		_LEAVE {
			if (scanpass == 2) {
				regstm(LEAVE);
				leavesp = stmcurr;
			}
		}
		;

redo:		_REDO {
			if (scanpass == 2) {
				regstm(REDO);
				redosp = stmcurr;
			}
		}
		;

next:		_NEXT {
			if (scanpass == 2) {
				regstm(NEXT);
				nextsp = stmcurr;
			}
		}
		;

skip:		_SKIP {
			if (scanpass == 2) {
				regstm(SKIP);
				skipsp = stmcurr;
			}
		}
		expnum
		;

last:		_LAST {
			if (scanpass == 2) {
				regstm(LAST);
				lastsp = stmcurr;
			}
		}
		;

pause:		_PAUSE {
			if (scanpass == 2) {
				regstm(PAUSE);
				pausesp = stmcurr;
			}
		} expnum
		;

quit:		_QUIT {
			if (scanpass == 2) {
				regstm(QUIT);
				quitsp = stmcurr;
			}
		}
		;

die:		_DIE {
			if (scanpass == 2) {
				regstm(DIE);
				quitsp = stmcurr;
			}
		} expstr
		;

output:		_OUTPUT {
			if (scanpass == 2) {
				regstm(OUTPUT);
				outputsp = stmcurr;
			}
		}
		expstr
		| _OUTPUT _MAIL {
			if (scanpass == 2) {
				regstm(MAIL);
				outputsp = stmcurr;
			}
		}
		expstr
		| _OUTPUT _PRINT {
			if (scanpass == 2) {
				regstm(PRINT);
				outputsp = stmcurr;
			}
		} expstr
		| _OUTPUT _LOGFILE {
			if (scanpass == 2) {
				regstm(LOCLOG);
				outputsp = stmcurr;
			}
		}
		_STR {
			if (scanpass == 2) {
				DUMP($4);
				stmcurr->logfile = $4;
			} else
				free($4);
		}
		expstr
		| _OUTPUT _PIKTLOG {
			if (scanpass == 2) {
				regstm(PIKTLOG);
				outputsp = stmcurr;
			}
		}
		expstr
		| _OUTPUT _SYSLOG {
			if (scanpass == 2) {
				regstm(SYSLOG);
				outputsp = stmcurr;
			}
		}
		expstr
		;

set:		_SET _VARNUM '[' {
			if (scanpass == 2) {
				DUMP($2);
				DUMP("[");
				regstm(SET);
				setsp = stmcurr;
				exprtyp = LHS;
				regtok(mystrdup("]",_altparse_), FALSE);
			}
			/* arytype, the 4th arg, and arydim, the 5th arg,
			   are really set after call to ndx following
			 */
			/* pass arndxnil instead
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = NILARNDX;
			 */
			v = regvar($2+1, valtype, NUMBER, NIL, NIL, arndxnil, (char *)NULL, (char *)NULL);
			free($2);
			arynum = 0;
		}
		ndx
		']' '=' {
			if (scanpass == 2) {
				regtok(mystrdup("[",_altparse_), FALSE);
				DUMP("]");
				exprtyp = RHS;
			}
			ARYTYP(v) = $5 & 0x01;
			if (isfakearray)
				ARYDIM(v) = 0;
			else
				ARYDIM(v) = $5>>1;
		}
		conditions
		| _SET _VARSTR '[' {
			if (scanpass == 2) {
				DUMP($2);
				DUMP("[");
				regstm(SET);
				setsp = stmcurr;
				exprtyp = LHS;
				regtok(mystrdup("]",_altparse_), FALSE);
			}
			/* artype, the 4th arg, and arydim, the 5th arg,
			   are really set after call to ndx following
			 */
			/* pass arndxnil instead
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = NILARNDX;
			 */
			v = regvar($2+1, valtype, STRING, NIL, NIL, arndxnil, (char *)NULL, (char *)NULL);
			free($2);
			arynum = 0;
		}
		ndx
		']' '=' {
			if (scanpass == 2) {
				DUMP("]");
				regtok(mystrdup("[",_altparse_), FALSE);
				exprtyp = RHS;
			}
			if (v >= 0) {	/* possibly v == OUTLINNOW */
				ARYTYP(v) = $5 & 0x01;
				if (isfakearray)
					ARYDIM(v) = 0;
				else
					ARYDIM(v) = $5>>1;
			}
		}
		expstr
		;

unset:		_UNSET varnow '[' {
			if (scanpass == 2) {
				DUMP($2);
				DUMP("[");
				regstm(UNSET);
				unsetsp = stmcurr;
				exprtyp = LHS;
				stmcurr->vrndx = varndx($2+1);
			}
			free($2);
		}
		expstr
		']' {
			if (scanpass == 2) {
				DUMP("]");
				exprtyp = RHS;
				/* these are just placeholders so that the
				   second exprparse() call in dostatement()
				   won't choke
				 */
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok(mystrdup(nilstr, _altparse_), FALSE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			}
		}
		;

exec:		_EXEC {
			if (scanpass == 2) {
				regstm(EXEC);
				execsp = stmcurr;
			}
		} expstr
		| _EXEC _WAIT {
			if (scanpass == 2) {
				regstm(EXECWAIT);
				execsp = stmcurr;
			}
		} expstr
		;

varstr:		_VARSTR {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARHSTSTR {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARPRCSTR {
			/* DUMP($1) */
			$$ = $1;
		}
		;

varnum:		_VARNUM {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARHSTNUM {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARPRCNUM {
			/* DUMP($1) */
			$$ = $1;
		}
		;

args:		/* empty */ { ; }
		| arg
		| argcoms arg
		;

argcoms:	argcom
		| argcoms argcom
		;

argcom:		arg
		',' {
			if (scanpass == 2) {
				regtok(mystrdup(",", _altparse_), FALSE);
			}
		}
		;

arg:		conditions
		| _FLHNDL {
			if (scanpass == 2) {
				regtok($1, FALSE);
			}
			/* whether FILEHANDLE or PROCHANDLE doesn't yet
			   matter at this point, so just pass FILEHANDLE
			 */
			/* pass arndxnil instead
			for (k=0; k<=MAXARYDIM; k++)
				arndx[k] = NILARNDX;
			 */
			regvar($1+1, valtype, FILEHANDLE, NIL, 0, arndxnil, (char *)NULL, (char *)NULL);
			if (scanpass == 1) {
				free($1);
			}
		}
		;

fnnum:		_FNNUM {
			nam = (char *) mymalloc((strlen($1+1)+3) * sizeof(char), _altparse_);
			sprintf(nam, " %s ", $1+1);
			if (strstr(" split parse members keys skeys sortkeys rkeys reversekeys ", nam))
				doregvar = TRUE;
			else
				doregvar = FALSE;
			if (scanpass == 2) {
				DUMP($1);
				regtok($1, TRUE);
			} else {
				if (ischeck || isdoublecheck) {
				 	if (! strstr(fnnumnames, nam)) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_NOT_NUMBER_FUNCTION, __FILE__, __LINE__, _altparse_, FALSE, "in %s, #%s() is not a number function", ALMNAM, $1);
						STATUS = ERROR;
					}
				}
				free($1);
			}
			free(nam);
		}
		'(' {
			if (scanpass == 2) {
				regtok(mystrdup("(", _altparse_), FALSE);
			}
		}
		args
		')' {
			if (scanpass == 2) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
			doregvar = FALSE;
		}
		;

fnstr:		_FNSTR {
			doregvar = FALSE;
			if (scanpass == 2) {
				DUMP($1);
				regtok($1, TRUE);
			} else {
				if (ischeck || isdoublecheck) {
					nam = (char *) mymalloc((strlen($1+1)+3) * sizeof(char), _altparse_);
					sprintf(nam, " %s ", $1+1);
					if (! strstr(fnstrnames, nam) &&
					    ! strstr(" inlin inline inputline ", nam)) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_NOT_STRING_FUNCTION, __FILE__, __LINE__, _altparse_, FALSE, "in %s, $%s() is not a string function", ALMNAM, $1);
						STATUS = ERROR;
					}
					free(nam);
				}
				free($1);
			}
		}
		'(' {
			if (scanpass == 2) {
				regtok(mystrdup("(", _altparse_), FALSE);
			}
		}
		args
		')' {
			if (scanpass == 2) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		;

ndx:		expstr {
			$$ = 1<<1 | STRING;	/*  1 dimension */
		}
		| {
			CHKARYNUM;
			if (arynum <= MAXARY-1)
				dim[arynum++] = 0;
		} ndxnums {
			arynum--;
			CHKARYNUM;
			$$ = $2<<1 | NUMBER;	/* $2 dimensions */
		}
		;

ndxnums:	expnum {
			$$ = ++dim[arynum-1];
		}
		| ndxnums ',' {
			if (scanpass == 2) {
				DUMP(",");
				regtok(mystrdup(",", _altparse_), FALSE);
			}
		} expnum {
			$$ = ++dim[arynum-1];
		}
		;

exp:		expstr { $$ = STRING; }
		| expnum { $$ = NUMBER; }
		;

expstr:		_STR {
			if (scanpass == 2) {
				DUMP($1);
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok($1, FALSE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			} else
				free($1);
		}
		| _STRVAR {
			if (scanpass == 2) {
				DUMP($1);
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok($1, TRUE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			} else
				free($1);
		}
		| _VALSTR {
			if (scanpass == 2) {
				DUMP($1);
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok($1, FALSE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			} else
				free($1);
		}
		| _VALSTR '[' {
			if (scanpass == 2) {
				DUMP($1);
				DUMP("[");
				switch ((vdt[vapndx] = STRING)) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
				}
				vapndx++;
			}
		}
		expnum
		']' {
			if (scanpass == 2) {
				vapndx--;
				DUMP("]");
				switch (vdt[vapndx]) {
					case STRING:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok(mystrdup("]", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| varstr '[' {
			if (scanpass == 2) {
				DUMP($1);
				DUMP("[");
				switch ((vdt[vapndx] = vardattype($1+1, (signed)strlen($1)-1, FALSE))) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
				}
				vapndx++;
			} else {
				if (doregvar) {	/* register first arg in
						   #split()/#parse()/
						   #members()/#keys()/
						   #skeys()/#sortkeys()/
						   #rkeys()/#reversekeys()
						 */
					/* pass arndxnil instead
					for (k=0; k<=MAXARYDIM; k++)
						arndx[k] = NILARNDX;
					 */
					v = regvar($1+1, valtype, STRING, NUMBER, 1, arndxnil, (char *)NULL, (char *)NULL);
					doregvar = FALSE;
				}
			}
		}
		ndx
		']' {
			if (scanpass == 2) {
				vapndx--;
				DUMP("]");
				switch (vdt[vapndx]) {
					case STRING:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok(mystrdup("]", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| fnstr
		| expstr
		_SOP {
			if (scanpass == 2) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expstr
		;

expnum:		_NUM {
			if (scanpass == 2) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		}
		| _INT {
			if (scanpass == 2) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		}
		| _VALNUM {
			if (scanpass == 2) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		}
		| varnum '[' {
			if (scanpass == 2) {
				DUMP($1);
				DUMP("[");
				switch ((vdt[vapndx] = vardattype($1+1, (signed)strlen($1)-1, FALSE))) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
				}
				vapndx++;
			}
		}
		ndx
		']' {
			if (scanpass == 2) {
				vapndx--;
				DUMP("]");
				switch (vdt[vapndx]) {
					case STRING:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok(mystrdup("]", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| fnnum
		| '(' {
			if (scanpass == 2) {
				regtok(mystrdup("(", _altparse_), FALSE);
			}
		} expnum ')' {
			if (scanpass == 2) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		| '+' {
			if (scanpass == 2) {
				regtok(mystrdup("+", _altparse_), FALSE);
			}
		}
		expnum %prec _UNOP
		| '-' {
			if (scanpass == 2) {
				regtok(mystrdup("-", _altparse_), FALSE);
			}
		}
		expnum %prec _UNOP
		| '~' {
			if (scanpass == 2) {
				regtok(mystrdup("~", _altparse_), FALSE);
			}
		}
		expnum
		| expnum
		_NOP {
			if (scanpass == 2) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expnum
		| expnum
		'+' {
			if (scanpass == 2) {
				regtok(mystrdup("+", _altparse_), FALSE);
			}
		}
		expnum
		| expnum
		'-' {
			if (scanpass == 2) {
				regtok(mystrdup("-", _altparse_), FALSE);
			}
		}
		expnum
		;

conditions:	expression
		| relation
		| '[' {
			if (scanpass == 2) {
				regtok(mystrdup("[", _altparse_), FALSE);
			}
		} conditions ')' {
			if (scanpass == 2) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		| _NOT {
			if (scanpass == 2) {
				regtok(mystrdup("!", _altparse_), FALSE);
			}
		} conditions
		| conditions _AND {
			if (scanpass == 2) {
				regtok(mystrdup("&&", _altparse_), FALSE);
			}
		} conditions
		| conditions _OR {
			if (scanpass == 2) {
				regtok(mystrdup("||", _altparse_), FALSE);
			}
		} conditions
		;

expression:	expnum
		| _FOP {
			if (scanpass == 2) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		} expstr
		| expstr
		;

relation:	expnum
		_NREL {
			if (scanpass == 2) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expnum
		| expstr
		_SREL {
			if (scanpass == 2) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expstr
		;

%%
